@@ -14,176 +14,103 @@
 # limitations under the License.
 #
 import time
-from typing import Any, Dict, Iterable, Optional, Tuple, Union
+from typing import Any, Iterable, Optional, Tuple, Union
 
-from .constants import *
-from .binary import GenericObjectMeta, unwrap_binary
 from .datatypes import prop_codes
 from .datatypes.internal import AnyDataObject
-from .exceptions import (
-    CacheCreationError, CacheError, ParameterError, SQLError,
-    connection_errors,
-)
-from .utils import (
-    cache_id, get_field_by_id, is_wrapped,
-    status_to_exception, unsigned
-)
+from .exceptions import CacheCreationError, CacheError, ParameterError, SQLError
+from .utils import cache_id, status_to_exception
 from .api.cache_config import (
-    cache_create, cache_create_with_config,
-    cache_get_or_create, cache_get_or_create_with_config,
-    cache_destroy, cache_get_configuration,
+    cache_create, cache_create_with_config, cache_get_or_create, cache_get_or_create_with_config, cache_destroy,
+    cache_get_configuration
 )
 from .api.key_value import (
-    cache_get, cache_put, cache_get_all, cache_put_all, cache_replace,
-    cache_clear, cache_clear_key, cache_clear_keys,
-    cache_contains_key, cache_contains_keys,
-    cache_get_and_put, cache_get_and_put_if_absent, cache_put_if_absent,
-    cache_get_and_remove, cache_get_and_replace,
-    cache_remove_key, cache_remove_keys, cache_remove_all,
-    cache_remove_if_equals, cache_replace_if_equals, cache_get_size,
+    cache_get, cache_put, cache_get_all, cache_put_all, cache_replace, cache_clear, cache_clear_key, cache_clear_keys,
+    cache_contains_key, cache_contains_keys, cache_get_and_put, cache_get_and_put_if_absent, cache_put_if_absent,
+    cache_get_and_remove, cache_get_and_replace, cache_remove_key, cache_remove_keys, cache_remove_all,
+    cache_remove_if_equals, cache_replace_if_equals, cache_get_size
 )
-from .api.sql import scan, scan_cursor_get_page, sql, sql_cursor_get_page
-from .api.affinity import cache_get_node_partitions
-
+from .cursors import ScanCursor, SqlCursor
 
 PROP_CODES = set([
     getattr(prop_codes, x)
     for x in dir(prop_codes)
     if x.startswith('PROP_')
 ])
-CACHE_CREATE_FUNCS = {
-    True: {
-        True: cache_get_or_create_with_config,
-        False: cache_create_with_config,
-    },
-    False: {
-        True: cache_get_or_create,
-        False: cache_create,
-    },
-}
 
 
-class Cache:
-    """
-    GridGain cache abstraction. Users should never use this class directly,
-    but construct its instances with
-    :py:meth:`~pygridgain.client.Client.create_cache`,
-    :py:meth:`~pygridgain.client.Client.get_or_create_cache` or
-    :py:meth:`~pygridgain.client.Client.get_cache` methods instead. See
-    :ref:`this example <create_cache>` on how to do it.
-    """
+def get_cache(client: 'Client', settings: Union[str, dict]) -> 'Cache':
+    name, settings = __parse_settings(settings)
+    if settings:
+        raise ParameterError('Only cache name allowed as a parameter')
 
-    affinity = None
-    _cache_id = None
-    _name = None
-    _client = None
-    _settings = None
+    return Cache(client, name)
 
-    @staticmethod
-    def _validate_settings(
-        settings: Union[str, dict] = None, get_only: bool = False,
-    ):
-        if any([
-            not settings,
-            type(settings) not in (str, dict),
-            type(settings) is dict and prop_codes.PROP_NAME not in settings,
-        ]):
-            raise ParameterError('You should supply at least cache name')
 
-        if all([
-            type(settings) is dict,
-            not set(settings).issubset(PROP_CODES),
-        ]):
+def create_cache(client: 'Client', settings: Union[str, dict]) -> 'Cache':
+    name, settings = __parse_settings(settings)
+
+    conn = client.random_node
+    if settings:
+        result = cache_create_with_config(conn, settings)
+    else:
+        result = cache_create(conn, name)
+
+    if result.status != 0:
+        raise CacheCreationError(result.message)
+
+    return Cache(client, name)
+
+
+def get_or_create_cache(client: 'Client', settings: Union[str, dict]) -> 'Cache':
+    name, settings = __parse_settings(settings)
+
+    conn = client.random_node
+    if settings:
+        result = cache_get_or_create_with_config(conn, settings)
+    else:
+        result = cache_get_or_create(conn, name)
+
+    if result.status != 0:
+        raise CacheCreationError(result.message)
+
+    return Cache(client, name)
+
+
+def __parse_settings(settings: Union[str, dict]) -> Tuple[Optional[str], Optional[dict]]:
+    if isinstance(settings, str):
+        return settings, None
+    elif isinstance(settings, dict) and prop_codes.PROP_NAME in settings:
+        name = settings[prop_codes.PROP_NAME]
+        if len(settings) == 1:
+            return name, None
+
+        if not set(settings).issubset(PROP_CODES):
             raise ParameterError('One or more settings was not recognized')
 
-        if get_only and type(settings) is dict and len(settings) != 1:
-            raise ParameterError('Only cache name allowed as a parameter')
+        return name, settings
+    else:
+        raise ParameterError('You should supply at least cache name')
 
-    def __init__(
-        self, client: 'Client', settings: Union[str, dict] = None,
-        with_get: bool = False, get_only: bool = False,
-    ):
-        """
-        Initialize cache object.
 
-        :param client: GridGain client,
-        :param settings: cache settings. Can be a string (cache name) or a dict
-         of cache properties and their values. In this case PROP_NAME is
-         mandatory,
-        :param with_get: (optional) do not raise exception, if the cache
-         is already exists. Defaults to False,
-        :param get_only: (optional) do not communicate with GridGain server
-         at all, only create Cache instance. Defaults to False.
-        """
+class BaseCache:
+    def __init__(self, client: 'BaseClient', name: str):
         self._client = client
-        self._validate_settings(settings)
-        if type(settings) == str:
-            self._name = settings
-        else:
-            self._name = settings[prop_codes.PROP_NAME]
-
-        if not get_only:
-            func = CACHE_CREATE_FUNCS[type(settings) is dict][with_get]
-            result = func(client.random_node, settings)
-            if result.status != 0:
-                raise CacheCreationError(result.message)
-
+        self._name = name
+        self._settings = None
         self._cache_id = cache_id(self._name)
-        self.affinity = {
-            'version': (0, 0),
-        }
-
-    def get_protocol_version(self) -> Optional[Tuple]:
-        """
-        Returns the tuple of major, minor, and revision numbers of the used
-        thin protocol version, or None, if no connection to the GridGain
-        cluster was not yet established.
-
-        This method is not a part of the public API. Unless you wish to
-        extend the `pygridgain` capabilities (with additional testing,
-        logging, examining connections, et c.) you probably should not use it.
-        """
-        return self.client.protocol_version
-
-    @property
-    def settings(self) -> Optional[dict]:
-        """
-        Lazy Cache settings. See the :ref:`example <sql_cache_read>`
-        of reading this property.
-
-        All cache properties are documented here: :ref:`cache_props`.
-
-        :return: dict of cache properties and their values.
-        """
-        if self._settings is None:
-            config_result = cache_get_configuration(
-                self.get_best_node(),
-                self._cache_id
-            )
-            if config_result.status == 0:
-                self._settings = config_result.value
-            else:
-                raise CacheError(config_result.message)
-
-        return self._settings
+        self._client.register_cache(self._cache_id)
 
     @property
     def name(self) -> str:
         """
-        Lazy cache name.
-
         :return: cache name string.
         """
-        if self._name is None:
-            self._name = self.settings[prop_codes.PROP_NAME]
-
         return self._name
 
     @property
-    def client(self) -> 'Client':
+    def client(self) -> 'BaseClient':
         """
-        GridGain :class:`~pygridgain.client.Client` object.
-
         :return: Client object, through which the cache is accessed.
         """
         return self._client
@@ -197,142 +124,57 @@ class Cache:
         """
         return self._cache_id
 
-    def _process_binary(self, value: Any) -> Any:
-        """
-        Detects and recursively unwraps Binary Object.
 
-        :param value: anything that could be a Binary Object,
-        :return: the result of the Binary Object unwrapping with all other data
-         left intact.
+class Cache(BaseCache):
+    """
+    GridGain cache abstraction. Users should never use this class directly,
+    but construct its instances with
+    :py:meth:`~pygridgain.client.Client.create_cache`,
+    :py:meth:`~pygridgain.client.Client.get_or_create_cache` or
+    :py:meth:`~pygridgain.client.Client.get_cache` methods instead. See
+    :ref:`this example <create_cache>` on how to do it.
+    """
+
+    def __init__(self, client: 'Client', name: str):
         """
-        if is_wrapped(value):
-            return unwrap_binary(self._client, value)
-        return value
+        Initialize cache object. For internal use.
+
+        :param client: Ignite client,
+        :param name: Cache name.
+        """
+        super().__init__(client, name)
+
+    def _get_best_node(self, key=None, key_hint=None):
+        return self.client.get_best_node(self._cache_id, key, key_hint)
+
+    @property
+    def settings(self) -> Optional[dict]:
+        """
+        Lazy Cache settings. See the :ref:`example <sql_cache_read>`
+        of reading this property.
+
+        All cache properties are documented here: :ref:`cache_props`.
+
+        :return: dict of cache properties and their values.
+        """
+        if self._settings is None:
+            config_result = cache_get_configuration(
+                self._get_best_node(),
+                self._cache_id
+            )
+            if config_result.status == 0:
+                self._settings = config_result.value
+            else:
+                raise CacheError(config_result.message)
+
+        return self._settings
 
     @status_to_exception(CacheError)
     def destroy(self):
         """
         Destroys cache with a given name.
         """
-        return cache_destroy(self.get_best_node(), self._cache_id)
-
-    @status_to_exception(CacheError)
-    def _get_affinity(self, conn: 'Connection') -> Dict:
-        """
-        Queries server for affinity mappings. Retries in case
-        of an intermittent error (most probably “Getting affinity for topology
-        version earlier than affinity is calculated”).
-
-        :param conn: connection to Igneite server,
-        :return: OP_CACHE_PARTITIONS operation result value.
-        """
-        for _ in range(AFFINITY_RETRIES or 1):
-            result = cache_get_node_partitions(conn, self._cache_id)
-            if result.status == 0 and result.value['partition_mapping']:
-                break
-            time.sleep(AFFINITY_DELAY)
-
-        return result
-
-    def get_best_node(
-        self, key: Any = None, key_hint: 'GridGainDataType' = None,
-    ) -> 'Connection':
-        """
-        Returns the node from the list of the nodes, opened by client, that
-        most probably contains the needed key-value pair. See IEP-23.
-
-        This method is not a part of the public API. Unless you wish to
-        extend the `pygridgain` capabilities (with additional testing,
-        logging, examining connections, et c.) you probably should not use it.
-
-        :param key: (optional) pythonic key,
-        :param key_hint: (optional) GridGain data type, for which the given
-         key should be converted,
-        :return: GridGain connection object.
-        """
-        conn = self._client.random_node
-
-        if self.client.partition_aware and key is not None:
-            if key_hint is None:
-                key_hint = AnyDataObject.map_python_type(key)
-
-            if self.affinity['version'] < self._client.affinity_version:
-                # update partition mapping
-                while True:
-                    try:
-                        self.affinity = self._get_affinity(conn)
-                        break
-                    except connection_errors:
-                        # retry if connection failed
-                        pass
-                    except CacheError:
-                        # server did not create mapping in time
-                        return conn
-
-                # flatten it a bit
-                try:
-                    self.affinity.update(self.affinity['partition_mapping'][0])
-                except IndexError:
-                    return conn
-                del self.affinity['partition_mapping']
-
-                # calculate the number of partitions
-                parts = 0
-                if 'node_mapping' in self.affinity:
-                    for p in self.affinity['node_mapping'].values():
-                        parts += len(p)
-
-                self.affinity['number_of_partitions'] = parts
-
-                for conn in self.client._nodes:
-                    if not conn.alive:
-                        conn.reconnect()
-            else:
-                # get number of partitions
-                parts = self.affinity.get('number_of_partitions')
-
-            if not parts:
-                return conn
-
-            if self.affinity['is_applicable']:
-                affinity_key_id = self.affinity['cache_config'].get(
-                    key_hint.type_id,
-                    None
-                )
-                if affinity_key_id and isinstance(key, GenericObjectMeta):
-                    key, key_hint = get_field_by_id(key, affinity_key_id)
-
-            # calculate partition for key or affinity key
-            # (algorithm is taken from `RendezvousAffinityFunction.java`)
-            base_value = key_hint.hashcode(key, self._client)
-            mask = parts - 1
-
-            if parts & mask == 0:
-                part = (base_value ^ (unsigned(base_value) >> 16)) & mask
-            else:
-                part = abs(base_value // parts)
-
-            assert 0 <= part < parts, 'Partition calculation has failed'
-
-            # search for connection
-            try:
-                node_uuid, best_conn = None, None
-                for u, p in self.affinity['node_mapping'].items():
-                    if part in p:
-                        node_uuid = u
-                        break
-
-                if node_uuid:
-                    for n in conn.client._nodes:
-                        if n.uuid == node_uuid:
-                            best_conn = n
-                            break
-                    if best_conn and best_conn.alive:
-                        conn = best_conn
-            except KeyError:
-                pass
-
-        return conn
+        return cache_destroy(self._get_best_node(), self._cache_id)
 
     @status_to_exception(CacheError)
     def get(self, key, key_hint: object = None) -> Any:
@@ -348,17 +190,17 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_get(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key,
             key_hint=key_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
     def put(
-        self, key, value, key_hint: object = None, value_hint: object = None
+            self, key, value, key_hint: object = None, value_hint: object = None
     ):
         """
         Puts a value with a given key to cache (overwriting existing value
@@ -375,7 +217,7 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_put(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id, key, value,
             key_hint=key_hint, value_hint=value_hint
         )
@@ -388,10 +230,10 @@ class Cache:
         :param keys: list of keys or tuples of (key, key_hint),
         :return: a dict of key-value pairs.
         """
-        result = cache_get_all(self.get_best_node(), self._cache_id, keys)
+        result = cache_get_all(self._get_best_node(), self._cache_id, keys)
         if result.value:
             for key, value in result.value.items():
-                result.value[key] = self._process_binary(value)
+                result.value[key] = self.client.unwrap_binary(value)
         return result
 
     @status_to_exception(CacheError)
@@ -404,11 +246,11 @@ class Cache:
          to save. Each key or value can be an item of representable
          Python type or a tuple of (item, hint),
         """
-        return cache_put_all(self.get_best_node(), self._cache_id, pairs)
+        return cache_put_all(self._get_best_node(), self._cache_id, pairs)
 
     @status_to_exception(CacheError)
     def replace(
-        self, key, value, key_hint: object = None, value_hint: object = None
+            self, key, value, key_hint: object = None, value_hint: object = None
     ):
         """
         Puts a value with a given key to cache only if the key already exist.
@@ -424,11 +266,11 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_replace(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id, key, value,
             key_hint=key_hint, value_hint=value_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
@@ -439,7 +281,7 @@ class Cache:
         :param keys: (optional) list of cache keys or (key, key type
          hint) tuples to clear (default: clear all).
         """
-        conn = self.get_best_node()
+        conn = self._get_best_node()
         if keys:
             return cache_clear_keys(conn, self._cache_id, keys)
         else:
@@ -458,12 +300,22 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_clear_key(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key,
             key_hint=key_hint
         )
 
+    @status_to_exception(CacheError)
+    def clear_keys(self, keys: Iterable):
+        """
+        Clears the cache key without notifying listeners or cache writers.
+
+        :param keys: a list of keys or (key, type hint) tuples
+        """
+
+        return cache_clear_keys(self._get_best_node(), self._cache_id, keys)
+
     @status_to_exception(CacheError)
     def contains_key(self, key, key_hint=None) -> bool:
         """
@@ -478,7 +330,7 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_contains_key(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key,
             key_hint=key_hint
@@ -492,7 +344,7 @@ class Cache:
         :param keys: a list of keys or (key, type hint) tuples,
         :return: boolean `True` when all keys are present, `False` otherwise.
         """
-        return cache_contains_keys(self._client, self._cache_id, keys)
+        return cache_contains_keys(self._get_best_node(), self._cache_id, keys)
 
     @status_to_exception(CacheError)
     def get_and_put(self, key, value, key_hint=None, value_hint=None) -> Any:
@@ -512,17 +364,17 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_get_and_put(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, value,
             key_hint, value_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
     def get_and_put_if_absent(
-        self, key, value, key_hint=None, value_hint=None
+            self, key, value, key_hint=None, value_hint=None
     ):
         """
         Puts a value with a given key to cache only if the key does not
@@ -540,12 +392,12 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_get_and_put_if_absent(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, value,
             key_hint, value_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
@@ -565,7 +417,7 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_put_if_absent(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, value,
             key_hint, value_hint
@@ -585,17 +437,17 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_get_and_remove(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key,
             key_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
     def get_and_replace(
-        self, key, value, key_hint=None, value_hint=None
+            self, key, value, key_hint=None, value_hint=None
     ) -> Any:
         """
         Puts a value with a given key to cache, returning previous value
@@ -614,12 +466,12 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_get_and_replace(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, value,
             key_hint, value_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
@@ -635,7 +487,7 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_remove_key(
-            self.get_best_node(key, key_hint), self._cache_id, key, key_hint
+            self._get_best_node(key, key_hint), self._cache_id, key, key_hint
         )
 
     @status_to_exception(CacheError)
@@ -647,7 +499,7 @@ class Cache:
         :param keys: list of keys or tuples of (key, key_hint) to remove.
         """
         return cache_remove_keys(
-            self.get_best_node(), self._cache_id, keys
+            self._get_best_node(), self._cache_id, keys
         )
 
     @status_to_exception(CacheError)
@@ -655,7 +507,7 @@ class Cache:
         """
         Removes all cache entries, notifying listeners and cache writers.
         """
-        return cache_remove_all(self.get_best_node(), self._cache_id)
+        return cache_remove_all(self._get_best_node(), self._cache_id)
 
     @status_to_exception(CacheError)
     def remove_if_equals(self, key, sample, key_hint=None, sample_hint=None):
@@ -674,7 +526,7 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         return cache_remove_if_equals(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, sample,
             key_hint, sample_hint
@@ -682,8 +534,8 @@ class Cache:
 
     @status_to_exception(CacheError)
     def replace_if_equals(
-        self, key, sample, value,
-        key_hint=None, sample_hint=None, value_hint=None
+            self, key, sample, value,
+            key_hint=None, sample_hint=None, value_hint=None
     ) -> Any:
         """
         Puts a value with a given key to cache only if the key already exists
@@ -704,31 +556,29 @@ class Cache:
             key_hint = AnyDataObject.map_python_type(key)
 
         result = cache_replace_if_equals(
-            self.get_best_node(key, key_hint),
+            self._get_best_node(key, key_hint),
             self._cache_id,
             key, sample, value,
             key_hint, sample_hint, value_hint
         )
-        result.value = self._process_binary(result.value)
+        result.value = self.client.unwrap_binary(result.value)
         return result
 
     @status_to_exception(CacheError)
-    def get_size(self, peek_modes=0):
+    def get_size(self, peek_modes=None):
         """
         Gets the number of entries in cache.
 
         :param peek_modes: (optional) limit count to near cache partition
          (PeekModes.NEAR), primary cache (PeekModes.PRIMARY), or backup cache
-         (PeekModes.BACKUP). Defaults to all cache partitions (PeekModes.ALL),
+         (PeekModes.BACKUP). Defaults to primary cache partitions (PeekModes.PRIMARY),
         :return: integer number of cache entries.
         """
         return cache_get_size(
-            self.get_best_node(), self._cache_id, peek_modes
+            self._get_best_node(), self._cache_id, peek_modes
         )
 
-    def scan(
-        self, page_size: int = 1, partitions: int = -1, local: bool = False
-    ):
+    def scan(self, page_size: int = 1, partitions: int = -1, local: bool = False) -> ScanCursor:
         """
         Returns all key-value pairs from the cache, similar to `get_all`, but
         with internal pagination, which is slower, but safer.
@@ -739,41 +589,15 @@ class Cache:
          (negative to query entire cache),
         :param local: (optional) pass True if this query should be executed
          on local node only. Defaults to False,
-        :return: generator with key-value pairs.
+        :return: Scan query cursor.
         """
-        node = self.get_best_node()
-
-        result = scan(
-            node,
-            self._cache_id,
-            page_size,
-            partitions,
-            local
-        )
-        if result.status != 0:
-            raise CacheError(result.message)
-
-        cursor = result.value['cursor']
-        for k, v in result.value['data'].items():
-            k = self._process_binary(k)
-            v = self._process_binary(v)
-            yield k, v
-
-        while result.value['more']:
-            result = scan_cursor_get_page(node, cursor)
-            if result.status != 0:
-                raise CacheError(result.message)
-
-            for k, v in result.value['data'].items():
-                k = self._process_binary(k)
-                v = self._process_binary(v)
-                yield k, v
+        return ScanCursor(self.client, self._cache_id, page_size, partitions, local)
 
     def select_row(
-        self, query_str: str, page_size: int = 1,
-        query_args: Optional[list] = None, distributed_joins: bool = False,
-        replicated_only: bool = False, local: bool = False, timeout: int = 0
-    ):
+            self, query_str: str, page_size: int = 1,
+            query_args: Optional[list] = None, distributed_joins: bool = False,
+            replicated_only: bool = False, local: bool = False, timeout: int = 0
+    ) -> SqlCursor:
         """
         Executes a simplified SQL SELECT query over data stored in the cache.
         The query returns the whole record (key and value).
@@ -790,46 +614,13 @@ class Cache:
          on local node only. Defaults to False,
         :param timeout: (optional) non-negative timeout value in ms. Zero
          disables timeout (default),
-        :return: generator with key-value pairs.
+        :return: Sql cursor.
         """
-        node = self.get_best_node()
-
-        def generate_result(value):
-            cursor = value['cursor']
-            more = value['more']
-            for k, v in value['data'].items():
-                k = self._process_binary(k)
-                v = self._process_binary(v)
-                yield k, v
-
-            while more:
-                inner_result = sql_cursor_get_page(node, cursor)
-                if result.status != 0:
-                    raise SQLError(result.message)
-                more = inner_result.value['more']
-                for k, v in inner_result.value['data'].items():
-                    k = self._process_binary(k)
-                    v = self._process_binary(v)
-                    yield k, v
-
         type_name = self.settings[
             prop_codes.PROP_QUERY_ENTITIES
         ][0]['value_type_name']
         if not type_name:
             raise SQLError('Value type is unknown')
-        result = sql(
-            node,
-            self._cache_id,
-            type_name,
-            query_str,
-            page_size,
-            query_args,
-            distributed_joins,
-            replicated_only,
-            local,
-            timeout
-        )
-        if result.status != 0:
-            raise SQLError(result.message)
 
-        return generate_result(result.value)
+        return SqlCursor(self.client, self._cache_id, type_name, query_str, page_size, query_args,
+                         distributed_joins, replicated_only, local, timeout)
