@@ -23,7 +23,6 @@ from .cache_config import (
 from .primitive import *
 from .standard import *
 
-
 __all__ = [
     'PropName', 'PropCacheMode', 'PropCacheAtomicityMode', 'PropBackupsNumber',
     'PropWriteSynchronizationMode', 'PropCopyOnRead', 'PropReadFromBackup',
@@ -68,7 +67,7 @@ def prop_map(code: int):
         PROP_CACHE_KEY_CONFIGURATION: PropCacheKeyConfiguration,
         PROP_DEFAULT_LOCK_TIMEOUT: PropDefaultLockTimeout,
         PROP_MAX_CONCURRENT_ASYNC_OPERATIONS: PropMaxConcurrentAsyncOperation,
-        PROP_PARTITION_LOSS_POLICY: PartitionLossPolicy,
+        PROP_PARTITION_LOSS_POLICY: PropPartitionLossPolicy,
         PROP_EAGER_TTL: PropEagerTTL,
         PROP_STATISTICS_ENABLED: PropStatisticsEnabled,
     }[code]
@@ -81,7 +80,7 @@ class PropBase:
     @classmethod
     def build_header(cls):
         return type(
-            cls.__name__+'Header',
+            cls.__name__ + 'Header',
             (ctypes.LittleEndianStructure,),
             {
                 '_pack_': 1,
@@ -111,11 +110,17 @@ class PropBase:
         stream.seek(init_pos + ctypes.sizeof(prop_class))
         return prop_class
 
+    @classmethod
+    async def parse_async(cls, stream):
+        return cls.parse(stream)
+
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
-        return cls.prop_data_class.to_python(
-            ctype_object.data, *args, **kwargs
-        )
+        return cls.prop_data_class.to_python(ctype_object.data, *args, **kwargs)
+
+    @classmethod
+    async def to_python_async(cls, ctype_object, *args, **kwargs):
+        return cls.to_python(ctype_object, *args, **kwargs)
 
     @classmethod
     def from_python(cls, stream, value):
@@ -125,6 +130,10 @@ class PropBase:
         stream.write(bytes(header))
         cls.prop_data_class.from_python(stream, value)
 
+    @classmethod
+    async def from_python_async(cls, stream, value):
+        return cls.from_python(stream, value)
+
 
 class PropName(PropBase):
     prop_code = PROP_NAME
