@@ -13,31 +13,34 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
+import asyncio
 from collections import OrderedDict
 import ctypes
 import decimal
 from datetime import date, datetime, timedelta
 from io import SEEK_CUR
-from typing import Any, Tuple, Union, Callable, List
+from typing import Any, Union, Callable, List
 import uuid
 
 import attr
 
-from pygridgain.constants import *
+from pygridgain.constants import PROTOCOL_BYTE_ORDER
 from pygridgain.exceptions import ParseError
 from pygridgain.utils import is_binary, is_hinted, is_iterable
 from .type_codes import *
 
 
 __all__ = [
-    'AnyDataArray', 'AnyDataObject', 'Struct', 'StructArray', 'tc_map',
-    'infer_from_python',
+    'AnyDataArray', 'AnyDataObject', 'Struct', 'StructArray', 'tc_map', 'infer_from_python', 'infer_from_python_async'
 ]
 
 from ..stream import READ_BACKWARD
 
 
-def tc_map(key: bytes, _memo_map: dict = {}):
+_tc_map = {}
+
+
+def tc_map(key: bytes):
     """
     Returns a default parser/generator class for the given type code.
 
@@ -50,7 +53,8 @@ def tc_map(key: bytes, _memo_map: dict = {}):
      of the “type code-type class” mapping,
     :return: parser/generator class for the type code.
     """
-    if not _memo_map:
+    global _tc_map
+    if not _tc_map:
         from pygridgain.datatypes import (
             Null, ByteObject, ShortObject, IntObject, LongObject, FloatObject,
             DoubleObject, CharObject, BoolObject, UUIDObject, DateObject,
@@ -65,7 +69,7 @@ def tc_map(key: bytes, _memo_map: dict = {}):
             MapObject, BinaryObject, WrappedDataObject,
         )
 
-        _memo_map = {
+        _tc_map = {
             TC_NULL: Null,
 
             TC_BYTE: ByteObject,
@@ -111,7 +115,7 @@ def tc_map(key: bytes, _memo_map: dict = {}):
             TC_COMPLEX_OBJECT: BinaryObject,
             TC_ARRAY_WRAPPED_OBJECTS: WrappedDataObject,
         }
-    return _memo_map[key]
+    return _tc_map[key]
 
 
 class Conditional:
@@ -124,11 +128,25 @@ class Conditional:
         self.var2 = var2
 
     def parse(self, stream, context):
-        return self.var1.parse(stream) if self.predicate1(context) else self.var2.parse(stream)
+        if self.predicate1(context):
+            return self.var1.parse(stream)
+        return self.var2.parse(stream)
+
+    async def parse_async(self, stream, context):
+        if self.predicate1(context):
+            return await self.var1.parse_async(stream)
+        return await self.var2.parse_async(stream)
 
     def to_python(self, ctype_object, context, *args, **kwargs):
-        return self.var1.to_python(ctype_object, *args, **kwargs) if self.predicate2(context)\
-            else self.var2.to_python(ctype_object, *args, **kwargs)
+        if self.predicate2(context):
+            return self.var1.to_python(ctype_object, *args, **kwargs)
+        return self.var2.to_python(ctype_object, *args, **kwargs)
+
+    async def to_python_async(self, ctype_object, context, *args, **kwargs):
+        if self.predicate2(context):
+            return await self.var1.to_python_async(ctype_object, *args, **kwargs)
+        return await self.var2.to_python_async(ctype_object, *args, **kwargs)
+
 
 @attr.s
 class StructArray:
@@ -139,7 +157,7 @@ class StructArray:
 
     def build_header_class(self):
         return type(
-            self.__class__.__name__+'Header',
+            self.__class__.__name__ + 'Header',
             (ctypes.LittleEndianStructure,),
             {
                 '_pack_': 1,
@@ -150,19 +168,34 @@ class StructArray:
         )
 
     def parse(self, stream):
-        counter_type_len = ctypes.sizeof(self.counter_type)
-        length = int.from_bytes(
-            stream.mem_view(offset=counter_type_len),
-            byteorder=PROTOCOL_BYTE_ORDER
-        )
-        stream.seek(counter_type_len, SEEK_CUR)
+        fields, length = [], self.__parse_length(stream)
 
-        fields = []
         for i in range(length):
             c_type = Struct(self.following).parse(stream)
             fields.append(('element_{}'.format(i), c_type))
 
-        data_class = type(
+        return self.__build_final_class(fields)
+
+    async def parse_async(self, stream):
+        fields, length = [], self.__parse_length(stream)
+
+        for i in range(length):
+            c_type = await Struct(self.following).parse_async(stream)
+            fields.append(('element_{}'.format(i), c_type))
+
+        return self.__build_final_class(fields)
+
+    def __parse_length(self, stream):
+        counter_type_len = ctypes.sizeof(self.counter_type)
+        length = int.from_bytes(
+            stream.slice(offset=counter_type_len),
+            byteorder=PROTOCOL_BYTE_ORDER
+        )
+        stream.seek(counter_type_len, SEEK_CUR)
+        return length
+
+    def __build_final_class(self, fields):
+        return type(
             'StructArray',
             (self.build_header_class(),),
             {
@@ -171,36 +204,47 @@ class StructArray:
             },
         )
 
-        return data_class
-
     def to_python(self, ctype_object, *args, **kwargs):
-        result = []
         length = getattr(ctype_object, 'length', 0)
-        for i in range(length):
-            result.append(
-                Struct(
-                    self.following, dict_type=dict
-                ).to_python(
-                    getattr(ctype_object, 'element_{}'.format(i)),
-                    *args, **kwargs
-                )
-            )
-        return result
+        return [
+            Struct(self.following, dict_type=dict).to_python(getattr(ctype_object, 'element_{}'.format(i)),
+                                                             *args, **kwargs)
+            for i in range(length)
+        ]
+
+    async def to_python_async(self, ctype_object, *args, **kwargs):
+        length = getattr(ctype_object, 'length', 0)
+        result_coro = [
+            Struct(self.following, dict_type=dict).to_python_async(getattr(ctype_object, 'element_{}'.format(i)),
+                                                                   *args, **kwargs)
+            for i in range(length)
+        ]
+        return await asyncio.gather(*result_coro)
 
     def from_python(self, stream, value):
-        length = len(value)
-        header_class = self.build_header_class()
-        header = header_class()
-        header.length = length
+        self.__write_header(stream, len(value))
 
-
-        stream.write(header)
-        for i, v in enumerate(value):
+        for v in value:
             for default_key, default_value in self.defaults.items():
                 v.setdefault(default_key, default_value)
             for name, el_class in self.following:
                 el_class.from_python(stream, v[name])
 
+    async def from_python_async(self, stream, value):
+        self.__write_header(stream, len(value))
+
+        for v in value:
+            for default_key, default_value in self.defaults.items():
+                v.setdefault(default_key, default_value)
+            for name, el_class in self.following:
+                await el_class.from_python_async(stream, v[name])
+
+    def __write_header(self, stream, length):
+        header_class = self.build_header_class()
+        header = header_class()
+        header.length = length
+        stream.write(header)
+
 
 @attr.s
 class Struct:
@@ -210,12 +254,7 @@ class Struct:
     defaults = attr.ib(type=dict, default={})
 
     def parse(self, stream):
-        fields, ctx = [], {}
-
-        for _, c_type in self.fields:
-            if isinstance(c_type, Conditional):
-                for name in c_type.fields:
-                    ctx[name] = None
+        fields, ctx = [], self.__prepare_conditional_ctx()
 
         for name, c_type in self.fields:
             is_cond = isinstance(c_type, Conditional)
@@ -224,7 +263,31 @@ class Struct:
             if name in ctx:
                 ctx[name] = stream.read_ctype(c_type, direction=READ_BACKWARD)
 
-        data_class = type(
+        return self.__build_final_class(fields)
+
+    async def parse_async(self, stream):
+        fields, ctx = [], self.__prepare_conditional_ctx()
+
+        for name, c_type in self.fields:
+            is_cond = isinstance(c_type, Conditional)
+            c_type = await c_type.parse_async(stream, ctx) if is_cond else await c_type.parse_async(stream)
+            fields.append((name, c_type))
+            if name in ctx:
+                ctx[name] = stream.read_ctype(c_type, direction=READ_BACKWARD)
+
+        return self.__build_final_class(fields)
+
+    def __prepare_conditional_ctx(self):
+        ctx = {}
+        for _, c_type in self.fields:
+            if isinstance(c_type, Conditional):
+                for name in c_type.fields:
+                    ctx[name] = None
+        return ctx
+
+    @staticmethod
+    def __build_final_class(fields):
+        return type(
             'StructLE',
             (ctypes.LittleEndianStructure,),
             {
@@ -233,11 +296,7 @@ class Struct:
             },
         )
 
-        return data_class
-
-    def to_python(
-        self, ctype_object, *args, **kwargs
-    ) -> Union[dict, OrderedDict]:
+    def to_python(self, ctype_object, *args, **kwargs) -> Union[dict, OrderedDict]:
         result = self.dict_type()
         for name, c_type in self.fields:
             is_cond = isinstance(c_type, Conditional)
@@ -251,13 +310,41 @@ class Struct:
             )
         return result
 
+    async def to_python_async(self, ctype_object, *args, **kwargs) -> Union[dict, OrderedDict]:
+        result = self.dict_type()
+        for name, c_type in self.fields:
+            is_cond = isinstance(c_type, Conditional)
+
+            if is_cond:
+                value = await c_type.to_python_async(
+                    getattr(ctype_object, name),
+                    result,
+                    *args, **kwargs
+                )
+            else:
+                value = await c_type.to_python_async(
+                    getattr(ctype_object, name),
+                    *args, **kwargs
+                )
+            result[name] = value
+        return result
+
     def from_python(self, stream, value):
-        for default_key, default_value in self.defaults.items():
-            value.setdefault(default_key, default_value)
+        self.__set_defaults(value)
 
         for name, el_class in self.fields:
             el_class.from_python(stream, value[name])
 
+    async def from_python_async(self, stream, value):
+        self.__set_defaults(value)
+
+        for name, el_class in self.fields:
+            await el_class.from_python_async(stream, value[name])
+
+    def __set_defaults(self, value):
+        for default_key, default_value in self.defaults.items():
+            value.setdefault(default_key, default_value)
+
 
 class AnyDataObject:
     """
@@ -266,6 +353,9 @@ class AnyDataObject:
     """
     _python_map = None
     _python_array_map = None
+    _map_obj_type = None
+    _collection_obj_type = None
+    _binary_obj_type = None
 
     @staticmethod
     def get_subtype(iterable, allow_none=False):
@@ -294,40 +384,57 @@ class AnyDataObject:
 
         # if an iterable contains items of more than one non-nullable type,
         # return None
-        if all([
-            isinstance(x, type_first)
-            or ((x is None) and allow_none) for x in iterator
-        ]):
+        if all(isinstance(x, type_first) or ((x is None) and allow_none) for x in iterator):
             return type_first
 
     @classmethod
     def parse(cls, stream):
-        type_code = bytes(stream.mem_view(offset=ctypes.sizeof(ctypes.c_byte)))
-        try:
-            data_class = tc_map(type_code)
-        except KeyError:
-            raise ParseError('Unknown type code: `{}`'.format(type_code))
+        data_class = cls.__data_class_parse(stream)
         return data_class.parse(stream)
 
+    @classmethod
+    async def parse_async(cls, stream):
+        data_class = cls.__data_class_parse(stream)
+        return await data_class.parse_async(stream)
+
+    @classmethod
+    def __data_class_parse(cls, stream):
+        type_code = stream.slice(offset=ctypes.sizeof(ctypes.c_byte))
+        try:
+            return tc_map(type_code)
+        except KeyError:
+            raise ParseError('Unknown type code: `{}`'.format(type_code))
+
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
+        data_class = cls.__data_class_from_ctype(ctype_object)
+        return data_class.to_python(ctype_object)
+
+    @classmethod
+    async def to_python_async(cls, ctype_object, *args, **kwargs):
+        data_class = cls.__data_class_from_ctype(ctype_object)
+        return await data_class.to_python_async(ctype_object)
+
+    @classmethod
+    def __data_class_from_ctype(cls, ctype_object):
         type_code = ctype_object.type_code.to_bytes(
             ctypes.sizeof(ctypes.c_byte),
             byteorder=PROTOCOL_BYTE_ORDER
         )
-        data_class = tc_map(type_code)
-        return data_class.to_python(ctype_object)
+        return tc_map(type_code)
 
     @classmethod
-    def _init_python_map(cls):
+    def _init_python_mapping(cls):
         """
         Optimizes Python types→GridGain types map creation for speed.
 
         Local imports seem inevitable here.
         """
         from pygridgain.datatypes import (
-            LongObject, DoubleObject, String, BoolObject, Null, UUIDObject,
-            DateObject, TimeObject, DecimalObject, ByteArrayObject,
+            LongObject, DoubleObject, String, BoolObject, Null, UUIDObject, DateObject, TimeObject,
+            DecimalObject, ByteArrayObject, LongArrayObject, DoubleArrayObject, StringArrayObject,
+            BoolArrayObject, UUIDArrayObject, DateArrayObject, TimeArrayObject, DecimalArrayObject,
+            MapObject, CollectionObject, BinaryObject
         )
 
         cls._python_map = {
@@ -345,17 +452,6 @@ class AnyDataObject:
             decimal.Decimal: DecimalObject,
         }
 
-    @classmethod
-    def _init_python_array_map(cls):
-        """
-        Optimizes  Python types→GridGain array types map creation for speed.
-        """
-        from pygridgain.datatypes import (
-            LongArrayObject, DoubleArrayObject, StringArrayObject,
-            BoolArrayObject, UUIDArrayObject, DateArrayObject, TimeArrayObject,
-            DecimalArrayObject,
-        )
-
         cls._python_array_map = {
             int: LongArrayObject,
             float: DoubleArrayObject,
@@ -369,18 +465,20 @@ class AnyDataObject:
             decimal.Decimal: DecimalArrayObject,
         }
 
+        cls._map_obj_type = MapObject
+        cls._collection_obj_type = CollectionObject
+        cls._binary_obj_type = BinaryObject
+
     @classmethod
     def map_python_type(cls, value):
-        from pygridgain.datatypes import (
-            MapObject, CollectionObject, BinaryObject,
-        )
-
-        if cls._python_map is None:
-            cls._init_python_map()
-        if cls._python_array_map is None:
-            cls._init_python_array_map()
+        if cls._python_map is None or cls._python_array_map is None:
+            cls._init_python_mapping()
 
         value_type = type(value)
+
+        if value_type in cls._python_map:
+            return cls._python_map[value_type]
+
         if is_iterable(value) and value_type not in (str, bytearray, bytes):
             value_subtype = cls.get_subtype(value)
             if value_subtype in cls._python_array_map:
@@ -393,7 +491,7 @@ class AnyDataObject:
                 isinstance(value[0], int),
                 isinstance(value[1], dict),
             ]):
-                return MapObject
+                return cls._map_obj_type
 
             if all([
                 value_subtype is None,
@@ -401,7 +499,7 @@ class AnyDataObject:
                 isinstance(value[0], int),
                 is_iterable(value[1]),
             ]):
-                return CollectionObject
+                return cls._collection_obj_type
 
             # no default for ObjectArrayObject, sorry
 
@@ -410,10 +508,8 @@ class AnyDataObject:
             )
 
         if is_binary(value):
-            return BinaryObject
+            return cls._binary_obj_type
 
-        if value_type in cls._python_map:
-            return cls._python_map[value_type]
         raise TypeError(
             'Type `{}` is invalid.'.format(value_type)
         )
@@ -423,6 +519,11 @@ class AnyDataObject:
         p_type = cls.map_python_type(value)
         p_type.from_python(stream, value)
 
+    @classmethod
+    async def from_python_async(cls, stream, value):
+        p_type = cls.map_python_type(value)
+        await p_type.from_python_async(stream, value)
+
 
 def infer_from_python(stream, value: Any):
     """
@@ -431,14 +532,26 @@ def infer_from_python(stream, value: Any):
     :param value: pythonic value or (value, type_hint) tuple,
     :return: bytes.
     """
-    if is_hinted(value):
-        value, data_type = value
-    else:
-        data_type = AnyDataObject
+    value, data_type = __unpack_hinted(value)
 
     data_type.from_python(stream, value)
 
 
+async def infer_from_python_async(stream, value: Any):
+    """
+    Async version of infer_from_python
+    """
+    value, data_type = __unpack_hinted(value)
+
+    await data_type.from_python_async(stream, value)
+
+
+def __unpack_hinted(value):
+    if is_hinted(value):
+        return value
+    return value, AnyDataObject
+
+
 @attr.s
 class AnyDataArray(AnyDataObject):
     """
@@ -448,7 +561,7 @@ class AnyDataArray(AnyDataObject):
 
     def build_header(self):
         return type(
-            self.__class__.__name__+'Header',
+            self.__class__.__name__ + 'Header',
             (ctypes.LittleEndianStructure,),
             {
                 '_pack_': 1,
@@ -459,16 +572,33 @@ class AnyDataArray(AnyDataObject):
         )
 
     def parse(self, stream):
-        header_class = self.build_header()
-        header = stream.read_ctype(header_class)
-        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
+        header, header_class = self.__parse_header(stream)
 
         fields = []
         for i in range(header.length):
             c_type = super().parse(stream)
             fields.append(('element_{}'.format(i), c_type))
 
-        final_class = type(
+        return self.__build_final_class(header_class, fields)
+
+    async def parse_async(self, stream):
+        header, header_class = self.__parse_header(stream)
+
+        fields = []
+        for i in range(header.length):
+            c_type = await super().parse_async(stream)
+            fields.append(('element_{}'.format(i), c_type))
+
+        return self.__build_final_class(header_class, fields)
+
+    def __parse_header(self, stream):
+        header_class = self.build_header()
+        header = stream.read_ctype(header_class)
+        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
+        return header, header_class
+
+    def __build_final_class(self, header_class, fields):
+        return type(
             self.__class__.__name__,
             (header_class,),
             {
@@ -476,34 +606,58 @@ class AnyDataArray(AnyDataObject):
                 '_fields_': fields,
             }
         )
-        return final_class
 
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
-        result = []
-        length = getattr(ctype_object, "length", None)
-        if length is None:
-            return None
-        for i in range(length):
-            result.append(
+        length = cls.__get_length(ctype_object)
+
+        return [
+            super().to_python(getattr(ctype_object, 'element_{}'.format(i)), *args, **kwargs)
+            for i in range(length)
+        ]
+
+    @classmethod
+    async def to_python_async(cls, ctype_object, *args, **kwargs):
+        length = cls.__get_length(ctype_object)
+
+        values = asyncio.gather(
+            *[
                 super().to_python(
                     getattr(ctype_object, 'element_{}'.format(i)),
                     *args, **kwargs
-                )
-            )
-        return result
+                ) for i in range(length)
+            ]
+        )
+        return await values
+
+    @staticmethod
+    def __get_length(ctype_object):
+        return getattr(ctype_object, "length", None)
 
     def from_python(self, stream, value):
-        header_class = self.build_header()
-        header = header_class()
-
         try:
             length = len(value)
         except TypeError:
             value = [value]
             length = 1
-        header.length = length
+        self.__write_header(stream, length)
 
-        stream.write(header)
         for x in value:
             infer_from_python(stream, x)
+
+    async def from_python_async(self, stream, value):
+        try:
+            length = len(value)
+        except TypeError:
+            value = [value]
+            length = 1
+        self.__write_header(stream, length)
+
+        for x in value:
+            await infer_from_python_async(stream, x)
+
+    def __write_header(self, stream, length):
+        header_class = self.build_header()
+        header = header_class()
+        header.length = length
+        stream.write(header)
