@@ -14,15 +14,35 @@
 # limitations under the License.
 #
 
-import attr
 import ctypes
+from io import SEEK_CUR
 from random import randint
 
+import attr
+
 from pygridgain.api.result import APIResult
-from pygridgain.connection import Connection
+from pygridgain.connection import Connection, AioConnection
 from pygridgain.constants import MIN_LONG, MAX_LONG, RHF_TOPOLOGY_CHANGED
-from pygridgain.queries.response import Response, SQLResponse
-from pygridgain.stream import BinaryStream, READ_BACKWARD
+from pygridgain.queries.response import Response
+from pygridgain.stream import AioBinaryStream, BinaryStream, READ_BACKWARD
+
+
+def query_perform(query_struct, conn, post_process_fun=None, **kwargs):
+    async def _async_internal():
+        result = await query_struct.perform_async(conn, **kwargs)
+        if post_process_fun:
+            return post_process_fun(result)
+        return result
+
+    def _internal():
+        result = query_struct.perform(conn, **kwargs)
+        if post_process_fun:
+            return post_process_fun(result)
+        return result
+
+    if isinstance(conn, AioConnection):
+        return _async_internal()
+    return _internal()
 
 
 @attr.s
@@ -30,6 +50,7 @@ class Query:
     op_code = attr.ib(type=int)
     following = attr.ib(type=list, factory=list)
     query_id = attr.ib(type=int, default=None)
+    response_type = attr.ib(type=type(Response), default=Response)
     _query_c_type = None
 
     @classmethod
@@ -49,32 +70,45 @@ class Query:
             )
         return cls._query_c_type
 
-    def _build_header(self, stream, values: dict):
+    def from_python(self, stream, values: dict = None):
+        init_pos, header = stream.tell(), self._build_header(stream)
+        values = values if values else None
+
+        for name, c_type in self.following:
+            c_type.from_python(stream, values[name])
+
+        self.__write_header(stream, header, init_pos)
+
+    async def from_python_async(self, stream, values: dict = None):
+        init_pos, header = stream.tell(), self._build_header(stream)
+        values = values if values else None
+
+        for name, c_type in self.following:
+            await c_type.from_python_async(stream, values[name])
+
+        self.__write_header(stream, header, init_pos)
+
+    def _build_header(self, stream):
         header_class = self.build_c_type()
         header_len = ctypes.sizeof(header_class)
-        init_pos = stream.tell()
-        stream.seek(init_pos + header_len)
+        stream.seek(header_len, SEEK_CUR)
 
         header = header_class()
         header.op_code = self.op_code
         if self.query_id is None:
             header.query_id = randint(MIN_LONG, MAX_LONG)
 
-        for name, c_type in self.following:
-            c_type.from_python(stream, values[name])
-
-        header.length = stream.tell() - init_pos - ctypes.sizeof(ctypes.c_int)
-        stream.seek(init_pos)
-
         return header
 
-    def from_python(self, stream, values: dict = None):
-        header = self._build_header(stream, values if values else {})
+    @staticmethod
+    def __write_header(stream, header, init_pos):
+        header.length = stream.tell() - init_pos - ctypes.sizeof(ctypes.c_int)
+        stream.seek(init_pos)
         stream.write(header)
 
     def perform(
         self, conn: Connection, query_params: dict = None,
-        response_config: list = None, sql: bool = False, **kwargs,
+        response_config: list = None, **kwargs,
     ) -> APIResult:
         """
         Perform query and process result.
@@ -84,26 +118,60 @@ class Query:
          Defaults to no parameters,
         :param response_config: (optional) response configuration − list of
          (name, type_hint) tuples. Defaults to empty return value,
-        :param sql: (optional) use normal (default) or SQL response class,
         :return: instance of :class:`~pygridgain.api.result.APIResult` with raw
          value (may undergo further processing in API functions).
         """
-        with BinaryStream(conn) as stream:
+        with BinaryStream(conn.client) as stream:
             self.from_python(stream, query_params)
-            conn.send(stream.getbuffer())
+            response_data = conn.request(stream.getvalue())
 
-        if sql:
-            response_struct = SQLResponse(protocol_version=conn.get_protocol_version(),
-                                          following=response_config, **kwargs)
-        else:
-            response_struct = Response(protocol_version=conn.get_protocol_version(),
-                                       following=response_config)
+        response_struct = self.response_type(protocol_context=conn.protocol_context,
+                                             following=response_config, **kwargs)
 
-        with BinaryStream(conn, conn.recv()) as stream:
+        with BinaryStream(conn.client, response_data) as stream:
             response_ctype = response_struct.parse(stream)
             response = stream.read_ctype(response_ctype, direction=READ_BACKWARD)
 
-        # this test depends on protocol version
+        result = self.__post_process_response(conn, response_struct, response)
+
+        if result.status == 0:
+            result.value = response_struct.to_python(response)
+        return result
+
+    async def perform_async(
+        self, conn: AioConnection, query_params: dict = None,
+        response_config: list = None, **kwargs,
+    ) -> APIResult:
+        """
+        Perform query and process result.
+
+        :param conn: connection to Ignite server,
+        :param query_params: (optional) dict of named query parameters.
+         Defaults to no parameters,
+        :param response_config: (optional) response configuration − list of
+         (name, type_hint) tuples. Defaults to empty return value,
+        :return: instance of :class:`~pygridgain.api.result.APIResult` with raw
+         value (may undergo further processing in API functions).
+        """
+        with AioBinaryStream(conn.client) as stream:
+            await self.from_python_async(stream, query_params)
+            data = await conn.request(stream.getvalue())
+
+        response_struct = self.response_type(protocol_context=conn.protocol_context,
+                                             following=response_config, **kwargs)
+
+        with AioBinaryStream(conn.client, data) as stream:
+            response_ctype = await response_struct.parse_async(stream)
+            response = stream.read_ctype(response_ctype, direction=READ_BACKWARD)
+
+        result = self.__post_process_response(conn, response_struct, response)
+
+        if result.status == 0:
+            result.value = await response_struct.to_python_async(response)
+        return result
+
+    @staticmethod
+    def __post_process_response(conn, response_struct, response):
         if getattr(response, 'flags', False) & RHF_TOPOLOGY_CHANGED:
             # update latest affinity version
             new_affinity = (response.affinity_version, response.affinity_minor)
@@ -113,10 +181,7 @@ class Query:
                 conn.client.affinity_version = new_affinity
 
         # build result
-        result = APIResult(response)
-        if result.status == 0:
-            result.value = response_struct.to_python(response)
-        return result
+        return APIResult(response)
 
 
 class ConfigQuery(Query):
@@ -143,7 +208,7 @@ class ConfigQuery(Query):
             )
         return cls._query_c_type
 
-    def _build_header(self, stream, values: dict):
-        header = super()._build_header(stream, values)
+    def _build_header(self, stream):
+        header = super()._build_header(stream)
         header.config_length = header.length - ctypes.sizeof(type(header))
         return header
