@@ -14,4 +14,6 @@
 # limitations under the License.
 #
 
-from .binary_stream import BinaryStream, READ_FORWARD, READ_BACKWARD
\ No newline at end of file
+from .binary_stream import BinaryStream, AioBinaryStream, READ_FORWARD, READ_BACKWARD
+
+__all__ = ['BinaryStream', 'AioBinaryStream', 'READ_BACKWARD', 'READ_FORWARD']
